/*
    SPDX-FileCopyrightText: 2009 Stephen Kelly <steveire@gmail.com>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include "kselectionproxymodeltestsuite.h"

#include "moc_kselectionproxymodeltestsuite.cpp"
